/// QasTools: Desktop toolset for the Linux sound system ALSA.
/// \copyright See COPYING file.

#include "mixer_simple.hpp"
#include "mwdg/inputs_setup.hpp"
#include "mwdg/mixer_device_setup.hpp"
#include "mwdg/mixer_sliders.hpp"
#include "mwdg/mixer_sliders_status_widget.hpp"
#include "mwdg/mixer_switches.hpp"
#include "qsnd/mixer/mixer.hpp"
#include "views/message_widget.hpp"
#include "views/mixer_simple_setup.hpp"
#include <QCoreApplication>
#include <QHBoxLayout>
#include <QKeyEvent>
#include <QStackedLayout>
#include <QVBoxLayout>
#include <QWidgetAction>
#include <iostream>

namespace Views
{

Mixer_Simple::Mixer_Simple ( Wdg::DS::Widget_Style_Db * wdg_style_db_n,
                             dpe::Image_Allocator * image_alloc_n,
                             QWidget * parent_n )
: Views::View_Base ( parent_n )
{
  // Mixer object
  {
    _qsnd_mixer = new QSnd::Mixer::Mixer ( this );
    connect ( _qsnd_mixer,
              &QSnd::Mixer::Mixer::sig_mixer_reload_request,
              this,
              &Mixer_Simple::sig_mdev_reload_request );
  }

  // Mixer sliders
  {
    _mixer_sliders = new MWdg::Mixer_Sliders ( wdg_style_db_n, image_alloc_n );

    connect ( _mixer_sliders,
              &MWdg::Mixer_Sliders::sig_footer_label_selected,
              this,
              &Mixer_Simple::footer_label_selected );
  }

  // Mixer switches
  {
    _mixer_switches = new MWdg::Mixer_Switches;
  }

  // Vertical splitter
  {
    _mixer_split.setOrientation ( Qt::Vertical );
    _mixer_split.addWidget ( _mixer_sliders );
    _mixer_split.addWidget ( _mixer_switches );

    QVBoxLayout * lay_vbox ( new QVBoxLayout );
    lay_vbox->setContentsMargins ( 0, 0, 0, 0 );
    lay_vbox->addWidget ( &_mixer_split );
    _wdg_areas.setLayout ( lay_vbox );

    lay_stack ()->addWidget ( &_wdg_areas );
  }

  // Adjust layout margins
  {
    QMargins mgs ( _mixer_sliders->contentsMargins () );
    mgs.setTop ( 0 );
    mgs.setBottom ( 0 );
    mgs.setLeft ( 0 );
    mgs.setRight ( 0 );
    _mixer_sliders->setContentsMargins ( mgs );
  }

  if ( _mixer_sliders->layout () != nullptr ) {
    QMargins mgs ( _mixer_sliders->layout ()->contentsMargins () );
    mgs.setTop ( 0 );
    mgs.setBottom ( 0 );
    mgs.setLeft ( 0 );
    mgs.setRight ( 0 );
    _mixer_sliders->layout ()->setContentsMargins ( mgs );
  }
}

Mixer_Simple::~Mixer_Simple ()
{
  set_mdev_setup ( nullptr );
  set_view_setup ( nullptr );
}

void
Mixer_Simple::set_mdev_setup ( const MWdg::Mixer_Device_Setup * setup_n )
{
  if ( mdev_setup () != nullptr ) {
    clear_view ();
    _qsnd_mixer->close ();
  }

  Views::View_Base::set_mdev_setup ( setup_n );

  if ( mdev_setup () != nullptr ) {
    _qsnd_mixer->open ( mdev_setup ()->ctl_addr );
    setup_view ();
  }
}

void
Mixer_Simple::set_inputs_setup ( const MWdg::Inputs_Setup * setup_n )
{
  Views::View_Base::set_inputs_setup ( setup_n );
  _mixer_sliders->set_inputs_setup ( setup_n );
  _mixer_switches->set_inputs_setup ( setup_n );
}

void
Mixer_Simple::set_view_setup ( Views::View_Base_Setup * setup_n )
{
  clear_view ();
  _view_setup = dynamic_cast< Views::Mixer_Simple_Setup * > ( setup_n );
  setup_view ();
}

void
Mixer_Simple::clear_view ()
{
  if ( ( mdev_setup () == nullptr ) || ( _view_setup == nullptr ) ) {
    return;
  }

  if ( _status_wdg != nullptr ) {
    delete _status_wdg;
    _status_wdg = nullptr;
  }

  if ( _qsnd_mixer->is_open () ) {
    _mixer_sliders->set_mixer_setup ( nullptr );
    _mixer_switches->set_mixer_setup ( nullptr );
  }

  emit sig_streams_available ( false );
}

void
Mixer_Simple::setup_view ()
{
  if ( ( mdev_setup () == nullptr ) || ( _view_setup == nullptr ) ) {
    return;
  }

  // Fix show stream flags on demand
  if ( !( _view_setup->show_stream[ 0 ] || _view_setup->show_stream[ 1 ] ) ) {
    _view_setup->show_stream[ 0 ] = true;
  }

  // The stream dirs shown are temporarily overridden for
  // devices with only one stream direction type
  {
    bool show_stream[ 2 ];
    show_stream[ 0 ] = _view_setup->show_stream[ 0 ];
    show_stream[ 1 ] = _view_setup->show_stream[ 1 ];
    if ( ( _qsnd_mixer->num_elems_stream ( 0 ) == 0 ) ||
         ( _qsnd_mixer->num_elems_stream ( 1 ) == 0 ) ) {
      show_stream[ 0 ] = ( _qsnd_mixer->num_elems_stream ( 0 ) > 0 );
      show_stream[ 1 ] = ( _qsnd_mixer->num_elems_stream ( 1 ) > 0 );
    }
    if ( ( _qsnd_mixer->num_elems_stream ( 0 ) > 0 ) &&
         ( _qsnd_mixer->num_elems_stream ( 1 ) > 0 ) ) {
      emit sig_streams_available ( true );
    }
    emit sig_show_playback_changed ( show_stream[ 0 ] );
    emit sig_show_capture_changed ( show_stream[ 1 ] );

    _smixer_setup.show_stream[ 0 ] = show_stream[ 0 ];
    _smixer_setup.show_stream[ 1 ] = show_stream[ 1 ];
  }
  _smixer_setup.show_slider_value_labels =
      _view_setup->show_slider_value_labels;
  _smixer_setup.mixer_simple = _qsnd_mixer;

  unsigned int lay_stack_idx = 0;
  if ( mdev_setup ()->ctl_addr.isEmpty () ) {
    message_wdg ()->set_no_device ();
  } else {
    if ( _qsnd_mixer->is_open () ) {
      lay_stack_idx = 1;
    } else {
      message_wdg ()->set_mixer_open_fail ( mdev_setup ()->ctl_addr,
                                            _qsnd_mixer->err_message (),
                                            _qsnd_mixer->err_func () );
    }
  }

  // Show simple mixer
  if ( lay_stack_idx == 1 ) {

    std::size_t num_vis_sliders = 0;
    std::size_t num_vis_switches = 0;

    for ( std::size_t ii = 0; ii < 2; ++ii ) {
      if ( _smixer_setup.show_stream[ ii ] ) {
        num_vis_sliders += _qsnd_mixer->num_elems_volume ( ii );
        num_vis_switches += _qsnd_mixer->num_elems_no_volume ( ii );
      }
    }

    bool vis_splitter = ( ( num_vis_sliders > 0 ) && ( num_vis_switches > 0 ) );

    // Hide areas before installing changes
    _mixer_split.hide ();
    _mixer_sliders->hide ();
    _mixer_switches->hide ();

    // Install mixer setup
    if ( num_vis_sliders > 0 ) {
      _mixer_sliders->set_mixer_setup ( &_smixer_setup );
    }
    if ( num_vis_switches > 0 ) {
      _mixer_switches->set_mixer_setup ( &_smixer_setup );
    }

    // Update splitter layout
    if ( vis_splitter ) {
      _mixer_split.addWidget ( _mixer_sliders );
      _mixer_split.addWidget ( _mixer_switches );
      _mixer_split.setCollapsible ( 0, false );
      _mixer_split.setCollapsible ( 1, false );
      _mixer_split.setStretchFactor ( 0, 1 );
      _mixer_split.setStretchFactor ( 1, 0 );

      _mixer_sliders->show ();
      _mixer_switches->show ();
      {
        // Shrink switches area to minimum size
        QList< int > sizes;
        sizes.append ( height () ); // A value way too big
        sizes.append ( 1 );         // A value way too small
        _mixer_split.setSizes ( sizes );
      }
      _mixer_split.show ();
    } else {
      if ( num_vis_sliders > 0 ) {
        _wdg_areas.layout ()->addWidget ( _mixer_sliders );
      } else {
        _wdg_areas.layout ()->addWidget ( _mixer_switches );
      }
      _mixer_sliders->setVisible ( num_vis_sliders > 0 );
      _mixer_switches->setVisible ( num_vis_switches > 0 );
    }
  }

  lay_stack ()->setCurrentIndex ( lay_stack_idx );
}

void
Mixer_Simple::show_playback ( bool flag_n )
{
  if ( _view_setup == nullptr ) {
    return;
  }
  if ( flag_n != _view_setup->show_stream[ 0 ] ) {
    clear_view ();

    _view_setup->show_stream[ 0 ] = flag_n;
    if ( !_view_setup->show_stream[ 0 ] && !_view_setup->show_stream[ 1 ] ) {
      _view_setup->show_stream[ 1 ] = true;
    }

    setup_view ();
    setFocus ();
  }
}

void
Mixer_Simple::show_capture ( bool flag_n )
{
  if ( _view_setup == nullptr ) {
    return;
  }
  if ( flag_n != _view_setup->show_stream[ 1 ] ) {
    clear_view ();

    _view_setup->show_stream[ 1 ] = flag_n;
    if ( !_view_setup->show_stream[ 0 ] && !_view_setup->show_stream[ 1 ] ) {
      _view_setup->show_stream[ 0 ] = true;
    }

    setup_view ();
    setFocus ();
  }
}

void
Mixer_Simple::toggle_show_playback ()
{
  if ( _view_setup != nullptr ) {
    show_playback ( !_view_setup->show_stream[ 0 ] );
  }
}

void
Mixer_Simple::toggle_show_capture ()
{
  if ( _view_setup != nullptr ) {
    show_capture ( !_view_setup->show_stream[ 1 ] );
  }
}

void
Mixer_Simple::show_slider_value_widget ()
{
  if ( _status_wdg == nullptr ) {
    MWdg::Mixer_Sliders_Status_Widget * swdg (
        new MWdg::Mixer_Sliders_Status_Widget ( this ) );
    swdg->setAttribute ( Qt::WA_DeleteOnClose );
    swdg->set_sliders_pad ( _mixer_sliders->sliders_pad () );
    swdg->slider_focus_changed ();

    _status_wdg = swdg;
    _status_wdg->show ();
  }
}

void
Mixer_Simple::footer_label_selected ( unsigned int group_idx_n,
                                      unsigned int column_idx_n )
{
  // std::cout << "Footer label selected " << group_idx_n << " " <<
  //  column_idx_n << "\n";
  bool keep_indices ( true );
  if ( _status_wdg == nullptr ) {
    show_slider_value_widget ();
  } else {
    if ( ( group_idx_n == _status_group_idx ) &&
         ( column_idx_n == _status_column_idx ) ) {
      _status_group_idx = ~0;
      _status_column_idx = ~0;
      _status_wdg->close ();
      keep_indices = false;
    }
  }

  if ( keep_indices ) {
    _status_group_idx = group_idx_n;
    _status_column_idx = column_idx_n;
  }
}

void
Mixer_Simple::keyPressEvent ( QKeyEvent * event_n )
{
  bool handled ( false );
  if ( inputs_setup () != nullptr ) {
    handled = true;
    const QKeySequence kseq ( event_n->key () );
    if ( kseq == inputs_setup ()->ks_toggle_vis_stream[ 0 ] ) {
      toggle_show_playback ();
    } else if ( kseq == inputs_setup ()->ks_toggle_vis_stream[ 1 ] ) {
      toggle_show_capture ();
    } else {
      handled = false;
    }
  }
  if ( !handled ) {
    Views::View_Base::keyPressEvent ( event_n );
  }
}

void
Mixer_Simple::showEvent ( QShowEvent * event_n )
{
  View_Base::showEvent ( event_n );
  if ( _status_wdg != nullptr ) {
    _status_wdg->show ();
  }
}

void
Mixer_Simple::hideEvent ( QHideEvent * event_n )
{
  View_Base::hideEvent ( event_n );
  if ( _status_wdg != nullptr ) {
    _status_wdg->hide ();
  }
}

} // namespace Views
