/// QasTools: Desktop toolset for the Linux sound system ALSA.
/// \copyright See COPYING file.

#include "mixer.hpp"
#include "qsnd/event_types.hpp"
#include "qsnd/mixer/elem.hpp"
#include "qsnd/mixer/filter.hpp"
#include <QCoreApplication>
#include <QEvent>
#include <iostream>
#include <sstream>

namespace QSnd::Mixer
{

Mixer::Mixer ( QObject * parent_n )
: QObject ( parent_n )
{
  _num_elems_stream[ 0 ] = 0;
  _num_elems_stream[ 1 ] = 0;
  _num_elems_volume[ 0 ] = 0;
  _num_elems_volume[ 1 ] = 0;
  _num_elems_no_volume[ 0 ] = 0;
  _num_elems_no_volume[ 1 ] = 0;
}

Mixer::~Mixer ()
{
  close ();
}

void
Mixer::append_elem_filter ( std::unique_ptr< Filter > filter_n )
{
  if ( filter_n ) {
    _elem_filters.push_back ( std::move ( filter_n ) );
  }
}

void
Mixer::close ()
{
  // Destroy socket notifiers
  _socket_notifiers.clear ();
  _pollfds.clear ();
  _mixer_elems.clear ();

  if ( snd_mixer () != nullptr ) {
    snd_mixer_set_callback ( snd_mixer (), nullptr );
    snd_mixer_set_callback_private ( snd_mixer (), nullptr );

    if ( _snd_hctl != nullptr ) {
      snd_mixer_detach_hctl ( snd_mixer (), _snd_hctl );
    }

    snd_mixer_close ( snd_mixer () );
    _snd_mixer = nullptr;
  }

  if ( _snd_hctl != nullptr ) {
    snd_hctl_close ( _snd_hctl );
    _snd_hctl = nullptr;
  }

  _num_elems_stream[ 0 ] = 0;
  _num_elems_stream[ 1 ] = 0;
  _num_elems_volume[ 0 ] = 0;
  _num_elems_volume[ 1 ] = 0;
  _num_elems_no_volume[ 0 ] = 0;
  _num_elems_no_volume[ 1 ] = 0;
}

int
Mixer::open ( const QString & dev_name_n )
{
  // std::cout << "Mixer::open " << dev_name_n.toLocal8Bit().constData()
  //<< "\n";

  close ();

  _err_func.clear ();
  _err_message.clear ();

  if ( dev_name_n.isEmpty () ) {
    _err_func = "open";
    _err_message = tr ( "Empty device name" );
    return -1;
  }

  int err;
  err = snd_mixer_open ( &_snd_mixer, 0 );
  if ( err < 0 ) {
    _err_func = "snd_mixer_open";
    _err_message = snd_error_qstring ( err );
    return err;
  }

  err = snd_mixer_selem_register ( _snd_mixer, NULL, NULL );
  if ( err < 0 ) {
    _err_func = "snd_mixer_selem_register";
    _err_message = snd_error_qstring ( err );
    close ();
    return err;
  }

  // Open control handle
  err = snd_hctl_open (
      &_snd_hctl, dev_name_n.toLocal8Bit ().constData (), SND_CTL_NONBLOCK );

  if ( err < 0 ) {
    _err_func = "snd_hctl_open";
    _err_message = snd_error_qstring ( err );
    close ();
    return err;
  }

  // Attach hctl handle to mixer
  err = snd_mixer_attach_hctl ( _snd_mixer, _snd_hctl );
  if ( err < 0 ) {
    _snd_hctl = 0;
    _err_func = "snd_mixer_attach_hctl";
    _err_message = snd_error_qstring ( err );
    close ();
    return err;
  }

  // Load mixer
  err = snd_mixer_load ( _snd_mixer );
  if ( err < 0 ) {
    _err_func = "snd_mixer_load";
    _err_message = snd_error_qstring ( err );
    close ();
    return err;
  }

  // Install alsa callback
  snd_mixer_set_callback_private ( snd_mixer (), this );
  snd_mixer_set_callback ( snd_mixer (), &Mixer::alsa_callback_mixer );

  // Create mixer elements
  err = create_mixer_elems ();
  if ( err < 0 ) {
    close ();
    return err;
  }

  // Create socket notifiers
  err = create_socket_notifiers ();
  if ( err < 0 ) {
    close ();
    return err;
  }

  return 0;
}

int
Mixer::create_mixer_elems ()
{
  // Create and Elem objects
  {
    snd_mixer_elem_t * elem = snd_mixer_first_elem ( snd_mixer () );
    while ( elem != nullptr ) {
      {
        // Allocate new element
        auto mx_elem = std::make_unique< Elem > ( this, snd_mixer (), elem );

        // Translated display name
        mx_elem->set_display_name ( QCoreApplication::translate (
            "ALSA::Elem_Name", mx_elem->elem_name () ) );

        // Add to list
        _mixer_elems.push_back ( std::move ( mx_elem ) );
      }
      elem = snd_mixer_elem_next ( elem );
    }
  }

  // Filter Elem objects
  if ( !_elem_filters.empty () ) {
    // List for dropped (filtered) elements
    Element_Vector elems_dropped;

    // Filter elements
    for ( auto & elem_filter : _elem_filters ) {
      elem_filter->filter ( _mixer_elems, elems_dropped );
    }
  }

  // Handle names that occur multiple times
  const QString str_new ( "%1, %2" );
  for ( std::size_t ii = 0; ii < _mixer_elems.size (); ++ii ) {
    const auto & mx_elem = _mixer_elems[ ii ];
    QString str_cur = mx_elem->display_name ();

    bool dbl = false;

    // Detect if there're doubles at all
    for ( std::size_t jj = ii + 1; jj < _mixer_elems.size (); ++jj ) {
      const auto & mx_elem2 = _mixer_elems[ jj ];
      if ( mx_elem2->display_name () == str_cur ) {
        dbl = true;
        break;
      }
    }

    // Rename doubles
    if ( dbl ) {
      for ( std::size_t jj = ii; jj < _mixer_elems.size (); ++jj ) {
        const auto & mx_elem2 = _mixer_elems[ jj ];
        if ( mx_elem2->display_name () == str_cur ) {
          QString nname ( str_new );
          nname = nname.arg ( str_cur );
          nname = nname.arg ( mx_elem2->elem_index () );
          mx_elem2->set_display_name ( nname );
        }
      }
    }
  }

  // Acquire some statistics
  for ( const auto & mx_elem : _mixer_elems ) {
    for ( std::size_t sdir = 0; sdir < 2; ++sdir ) {

      bool has_vol = mx_elem->has_volume ( sdir );
      bool has_switch = mx_elem->has_switch ( sdir );
      bool has_enum = mx_elem->has_enum ( sdir );

      if ( has_vol ) {
        ++_num_elems_volume[ sdir ];
      } else {
        if ( has_switch || has_enum ) {
          ++_num_elems_no_volume[ sdir ];
        }
      }
      if ( has_vol || has_switch || has_enum ) {
        ++_num_elems_stream[ sdir ];
      }
    }
  }

  return 0;
}

int
Mixer::create_socket_notifiers ()
{
  // std::cout << "Mixer::create_socket_notifiers" << "\n";

  // Setup socket notifiers
  {
    int num_fds = snd_mixer_poll_descriptors_count ( _snd_mixer );
    if ( num_fds >= 0 ) {
      _pollfds.resize ( num_fds );
    } else {
      _err_func = "snd_mixer_poll_descriptors_count";
      _err_message = snd_error_qstring ( num_fds );
      return num_fds;
    }
  }

  if ( _pollfds.size () > 0 ) {
    memset ( _pollfds.data (), 0, _pollfds.size () * sizeof ( pollfd ) );
    int err = snd_mixer_poll_descriptors (
        _snd_mixer, _pollfds.data (), _pollfds.size () );
    if ( err < 0 ) {
      _err_func = "snd_mixer_poll_descriptors";
      _err_message = snd_error_qstring ( err );
      return err;
    }
  }

  for ( const auto & pfd : _pollfds ) {
    int fd = pfd.fd;
    if ( fd != 0 ) {
      auto notifier = std::make_unique< QSocketNotifier > (
          fd, QSocketNotifier::Read, this );
      connect ( notifier.get (),
                &QSocketNotifier::activated,
                this,
                &Mixer::socket_event );
      _socket_notifiers.push_back ( std::move ( notifier ) );
    }
  }

  return 0;
}

void
Mixer::set_socked_notifiers_enabled ( bool flag_n )
{
  for ( auto & notifier : _socket_notifiers ) {
    notifier->setEnabled ( flag_n );
  }
}

void
Mixer::signalize_all_changes ()
{
  for ( const auto & mx_elem : _mixer_elems ) {
    mx_elem->signalize_changes ();
  }
}

bool
Mixer::event ( QEvent * event_n )
{
  bool res = true;

  if ( event_n->type () == QSnd::evt_update_values_request ) {

    if ( !_update_requested ) {
      _update_requested = true;
      QCoreApplication::postEvent ( this,
                                    new QEvent ( QSnd::evt_update_values ) );
    }

  } else if ( event_n->type () == QSnd::evt_update_values ) {

    _update_requested = false;
    signalize_all_changes ();

  } else if ( event_n->type () == QSnd::evt_reload_request ) {

    if ( !_reload_requested ) {
      _reload_requested = true;
      QCoreApplication::postEvent ( this, new QEvent ( QSnd::evt_reload ) );
    }

  } else if ( event_n->type () == QSnd::evt_reload ) {

    _reload_requested = false;
    emit sig_mixer_reload_request ();

  } else {

    res = QObject::event ( event_n );
  }

  return res;
}

void
Mixer::socket_event ( int )
{
  if ( snd_mixer () == nullptr ) {
    return;
  }

  int num_ev = snd_mixer_handle_events ( snd_mixer () );
  if ( num_ev < 0 ) {
    // Error. Disable socket notifiers
    set_socked_notifiers_enabled ( false );
    print_alsa_error ( "snd_mixer_handle_events ", num_ev );
  }

  signalize_all_changes ();
}

void
Mixer::request_reload ()
{
  QCoreApplication::postEvent ( this, new QEvent ( QSnd::evt_reload_request ) );
}

// Alsa callbacks

int
Mixer::alsa_callback_mixer ( snd_mixer_t * snd_mixer_n,
                             unsigned int mask_n,
                             snd_mixer_elem_t * )
{
  int res = 0;

  Mixer * qsnd_mixer = nullptr;
  {
    void * priv ( snd_mixer_get_callback_private ( snd_mixer_n ) );
    qsnd_mixer = reinterpret_cast< Mixer * > ( priv );
  }

  if ( qsnd_mixer != nullptr ) {
    const unsigned int change_mask (
        SND_CTL_EVENT_MASK_VALUE | SND_CTL_EVENT_MASK_INFO |
        SND_CTL_EVENT_MASK_ADD | SND_CTL_EVENT_MASK_TLV );

    if ( ( mask_n == SND_CTL_EVENT_MASK_REMOVE ) ||
         ( ( mask_n & change_mask ) != 0 ) ) {
      qsnd_mixer->request_reload ();
    } else {
      {
        std::stringstream msg;
        msg << "Mixer::alsa_callback_mixer: ";
        msg << "Unknown mask ( " << mask_n << " )" << std::endl;
        std::cerr << msg.str ();
      }
      res = -1;
    }
  }

  return res;
}

} // namespace QSnd::Mixer
