/// QasTools: Desktop toolset for the Linux sound system ALSA.
/// \copyright See COPYING file.

#pragma once

#include "qsnd/alsa.hpp"
#include <QObject>
#include <QSocketNotifier>
#include <memory>
#include <vector>

namespace QSnd::Mixer
{

// Forward declaration
class Elem;
class Filter;

/// @brief Mixer_Simple
///
/// Connects ALSA simple mixer objects with Qt objects
///
class Mixer : public QObject
{
  Q_OBJECT

  public:
  // -- Public types

  using Element_Vector = std::vector< std::unique_ptr< QSnd::Mixer::Elem > >;

  public:
  // -- Construction

  Mixer ( QObject * parent_n = nullptr );

  ~Mixer ();

  // -- ALSA structs

  snd_hctl_t *
  snd_hctl () const
  {
    return _snd_hctl;
  }

  snd_mixer_t *
  snd_mixer () const
  {
    return _snd_mixer;
  }

  // -- Elements

  std::size_t
  num_elems () const
  {
    return _mixer_elems.size ();
  }

  /// @brief Number of elements in playback (0) or capture (1)
  std::size_t
  num_elems_stream ( std::size_t stream_dir_n ) const
  {
    return _num_elems_stream[ stream_dir_n ];
  }

  /// @brief Elements with a volume
  std::size_t
  num_elems_volume ( std::size_t stream_dir_n ) const
  {
    return _num_elems_volume[ stream_dir_n ];
  }

  /// @brief Elements with no volume
  std::size_t
  num_elems_no_volume ( std::size_t stream_dir_n ) const
  {
    return _num_elems_no_volume[ stream_dir_n ];
  }

  const Elem *
  elem ( std::size_t idx_n ) const
  {
    return _mixer_elems[ idx_n ].get ();
  }

  Elem *
  elem ( std::size_t idx_n )
  {
    return _mixer_elems[ idx_n ].get ();
  }

  // -- Element filters

  /// @brief Appends a new element filter
  ///
  /// The ownership of the filter is passed to the mixer.
  void
  append_elem_filter ( std::unique_ptr< Filter > filter_n );

  // -- open / close

  /// @brief Opens a simple mixer
  ///
  /// @return On an error a negative error code
  int
  open ( const QString & dev_name_n );

  void
  close ();

  bool
  is_open () const
  {
    return ( _snd_mixer != nullptr );
  }

  // -- Error strings / codes

  const QString &
  err_func () const
  {
    return _err_func;
  }

  const QString &
  err_message () const
  {
    return _err_message;
  }

  // -- Event handling

  bool
  event ( QEvent * event_n ) override;

  // -- Alsa callbacks

  static int
  alsa_callback_mixer ( snd_mixer_t * snd_mixer_n,
                        unsigned int mask_n,
                        snd_mixer_elem_t * );

  // -- Signals

  Q_SIGNAL
  void
  sig_mixer_reload_request ();

  // -- Public slots

  Q_SLOT
  void
  request_reload ();

  // Private slots
  private:
  Q_SLOT
  void
  socket_event ( int socket_id );

  private:
  // -- Private methods

  // Mixer elements

  int
  create_mixer_elems ();

  // Socket notifiers

  int
  create_socket_notifiers ();

  void
  set_socked_notifiers_enabled ( bool flag_n );

  void
  signalize_all_changes ();

  private:
  // -- Attributes
  snd_hctl_t * _snd_hctl = nullptr;
  snd_mixer_t * _snd_mixer = nullptr;

  Element_Vector _mixer_elems;
  std::vector< pollfd > _pollfds;
  std::vector< std::unique_ptr< QSocketNotifier > > _socket_notifiers;

  // Statistics
  std::size_t _num_elems_stream[ 2 ];
  std::size_t _num_elems_volume[ 2 ];    // Elements with a volume
  std::size_t _num_elems_no_volume[ 2 ]; // Elements without a volume

  bool _update_requested = false;
  bool _reload_requested = false;

  QString _err_func;
  QString _err_message;

  std::vector< std::unique_ptr< Filter > > _elem_filters;
};

} // namespace QSnd::Mixer
