/// QasTools: Desktop toolset for the Linux sound system ALSA.
/// \copyright See COPYING file.

#include "mixers_model.hpp"
#include "qsnd/cards_model.hpp"
#include <iostream>

namespace QSnd
{

Mixers_Model::Mixers_Model ( QObject * parent_n )
: QAbstractListModel ( parent_n )
, _cards_model ( std::make_unique< QSnd::Cards_Model > ( this ) )
{
  _default_info.acquire_info ( "default" );

  connect ( _cards_model.get (),
            &QAbstractItemModel::rowsInserted,
            this,
            &Mixers_Model::cards_inserted );

  connect ( _cards_model.get (),
            &QAbstractItemModel::rowsRemoved,
            this,
            &Mixers_Model::cards_removed );
}

Mixers_Model::~Mixers_Model () = default;

int
Mixers_Model::count () const
{
  return _cards_model->count () + 2;
}

QModelIndex
Mixers_Model::control_index ( const QSnd::CTL_Address & ctl_addr_n ) const
{
  // Default mixer
  if ( ctl_addr_n.addr_str () == "default" ) {
    return index ( 0, 0 );
  }
  // Card
  if ( ( ctl_addr_n.ctl_name () == "hw" ) && ( ctl_addr_n.num_args () == 1 ) ) {
    const auto & num_arg = ctl_addr_n.arg ( 0 );
    if ( ( num_arg.name ().isEmpty () || num_arg.name () == "CARD" ) ) {
      bool ok = false;
      int card_index = num_arg.value ().toInt ( &ok );
      if ( ok ) {
        for ( std::size_t row = 0; row < _cards_model->num_cards (); ++row ) {
          const auto & handle = _cards_model->card_info ( row );
          if ( handle->index () == card_index ) {
            return index ( 1 + row, 0 );
          }
        }
      }
    }
  }
  // User defined
  return index ( count () - 1, 0 );
}

Mixers_Model::Entry
Mixers_Model::comp_entry ( const QModelIndex & index_n ) const
{
  if ( ( !index_n.isValid () ) || ( index_n.parent ().isValid () ) ||
       ( index_n.column () != 0 ) || ( index_n.row () < 0 ) ||
       ( index_n.row () >= count () ) ) {
    return ENTRY_INVALID;
  }
  if ( index_n.row () == 0 ) {
    return ENTRY_DEFAULT;
  }
  if ( index_n.row () <= _cards_model->count () ) {
    return ENTRY_CARD;
  }
  return ENTRY_USER_DEFINED;
}

std::shared_ptr< const QSnd::Card_Info >
Mixers_Model::card_info ( const QModelIndex & index_n ) const
{
  Entry entry = comp_entry ( index_n );
  if ( entry == ENTRY_CARD ) {
    return _cards_model->card_info ( index_n.row () - 1 );
  }
  return {};
}

void
Mixers_Model::reload ()
{
  reload_begin ();
  reload_finish ();
}

void
Mixers_Model::reload_begin ()
{
  beginResetModel ();
}

void
Mixers_Model::reload_finish ()
{
  _default_info.acquire_info ( "default" );
  _cards_model->reload ();
  endResetModel ();

  emit countChanged ();
}

void
Mixers_Model::cards_inserted ( const QModelIndex & parent_n,
                               int start_n,
                               int end_n )
{
  if ( parent_n.isValid () ) {
    return;
  }
  beginInsertRows ( QModelIndex (), start_n + 1, end_n + 1 );
  endInsertRows ();
}

void
Mixers_Model::cards_removed ( const QModelIndex & parent_n,
                              int start_n,
                              int end_n )
{
  if ( parent_n.isValid () ) {
    return;
  }
  beginRemoveRows ( QModelIndex (), start_n + 1, end_n + 1 );
  endRemoveRows ();
}

QHash< int, QByteArray >
Mixers_Model::roleNames () const
{
  auto res = QAbstractListModel::roleNames ();
  res.insert ( ROLE_INDEX, "index" );
  res.insert ( ROLE_ID, "id" );
  res.insert ( ROLE_DRIVER, "driver" );
  res.insert ( ROLE_NAME, "name" );
  res.insert ( ROLE_LONG_NAME, "longName" );
  res.insert ( ROLE_MIXER_NAME, "mixerName" );
  res.insert ( ROLE_COMPONENTS, "components" );
  return res;
}

int
Mixers_Model::rowCount ( const QModelIndex & parent_n ) const
{
  if ( parent_n.isValid () ) {
    return 0;
  }
  return count ();
}

QVariant
Mixers_Model::data ( const QModelIndex & index_n, int role_n ) const
{
  Entry entry = comp_entry ( index_n );
  const QSnd::Card_Info * handle = nullptr;
  switch ( entry ) {
  case ENTRY_INVALID:
    return QVariant ();
  case ENTRY_DEFAULT:
    handle = &_default_info;
    break;
  case ENTRY_CARD:
    handle = _cards_model->card_info ( index_n.row () - 1 ).get ();
    break;
  case ENTRY_USER_DEFINED:
    handle = &_user_defined_info;
    break;
  }

  if ( handle != nullptr ) {
    switch ( role_n ) {
    case Qt::DisplayRole:
      switch ( entry ) {
      case ENTRY_INVALID:
        return QVariant ();
      case ENTRY_DEFAULT:
        return QVariant ( tr ( "Default" ) );
      case ENTRY_CARD:
        return QVariant ( handle->name () );
      case ENTRY_USER_DEFINED:
        return QVariant ( tr ( "User defined" ) );
      }
      break;
    case ROLE_INDEX:
      return QVariant ( handle->index () );
    case ROLE_ID:
      return QVariant ( handle->id () );
    case ROLE_DRIVER:
      return QVariant ( handle->driver () );
    case ROLE_NAME:
      return QVariant ( handle->name () );
    case ROLE_LONG_NAME:
      return QVariant ( handle->long_name () );
    case ROLE_MIXER_NAME:
      return QVariant ( handle->mixer_name () );
    case ROLE_COMPONENTS:
      return QVariant ( handle->components () );
    default:
      break;
    }
  }
  return QVariant ();
}

} // namespace QSnd
